import time
import typing

# ROS2
import builtin_interfaces
import rclpy
from rclpy.node import Node

from sensor_msgs.msg import Image

# Ignition
from ignition.msgs.camera_info_pb2 import CameraInfo
from ignition.msgs import image_pb2 as ign_image
from ignition.msgs.image_pb2 import Image as IgnImage
from ignition.transport import Node as IgnNode
from ignition.transport import SubscribeOptions

# Camera publishers class
class CameraPublishers(object):
    def __init__(self, ros2_node, world_name, model_name, sensor_name):
        super().__init__()
        prefix = model_name + "/" + sensor_name
        self.ros2_node = ros2_node

        self.world_name = world_name
        self.model_name = model_name
        self.sensor_name = sensor_name

        self.image_publisher = self.ros2_node.create_publisher(
            Image, prefix + "/image", 10)

    def pub_image(self, value):
        self.image_publisher.publish(value)


class CameraSubscribers(object):
    def __init__(self, ros2_node, world_name, model_name, sensor_name):
        super().__init__()

        self.ros2_node = ros2_node

        self.world_name = world_name
        self.model_name = model_name
        self.sensor_name = sensor_name
        
        self.ign_node = IgnNode()
        self.last_image = IgnImage()

        self._sub_to_image(world_name, model_name, sensor_name)

    def _subscribe(self, topic, callback, msg_type_name, opts):
        if self.ign_node.subscribe(topic, callback, msg_type_name, opts):
            self.ros2_node.get_logger().info(
                "Subscribing to type [{}] on topic [{}]".format(
                    msg_type_name, topic))
        else:
            self.ros2_node.get_logger().error(
                "Error subscribing to topic [{}]".format(topic))

    def _sub_to_image(self, world_name, model_name, sensor_name):
        camera_topic = "{}".format(
            sensor_name)

        msg_type_name = IgnImage.DESCRIPTOR.full_name
        opts = SubscribeOptions()

        # callbacks to capture the image msgs
        def camera_cb(msg: IgnImage) -> None:
            self.last_image = msg

        self._subscribe(camera_topic, camera_cb, msg_type_name, opts)


class PixelFormatType(object):
    '''
    Pixel format code conversion.
    '''

    def __init__(self):
        super().__init_()

    def ign_to_ros(format):
        '''
        Convert Ignition pixel format code to ROS
        
        https://github.com/ignitionrobotics/ros_ign/blob/melodic/ros_ign_bridge/src/convert.cpp
    
        http://docs.ros.org/en/jade/api/sensor_msgs/html/image__encodings_8h_source.html
        '''

        ros_encoding = ""
        num_channels = 0
        octets_per_channel = 0

        if format is ign_image.UNKNOWN_PIXEL_FORMAT:
            ros_encoding = "unknown"
        elif format is ign_image.L_INT8:
            ros_encoding = "mono8"
            num_channels = 1
            octets_per_channel = 1
        elif format is ign_image.L_INT16:
            ros_encoding = "mono16"
            num_channels = 1
            octets_per_channel = 2
        elif format is ign_image.RGB_INT8:
            ros_encoding = "rgb8"
            num_channels = 3
            octets_per_channel = 1
        elif format is ign_image.RGBA_INT8:
            ros_encoding = "rgba8"
            num_channels = 4
            octets_per_channel = 1
        elif format is ign_image.BGRA_INT8:
            ros_encoding = "bgra8"
            num_channels = 4
            octets_per_channel = 1
        elif format is ign_image.RGB_INT16:
            ros_encoding = "rgb16"
            num_channels = 3
            octets_per_channel = 2
        elif format is ign_image.RGB_INT32:
            ros_encoding = "unknown"
            num_channels = 3
            octets_per_channel = 4
        elif format is ign_image.BGR_INT8:
            ros_encoding = "bgr8"
            num_channels = 3
            octets_per_channel = 1
        elif format is ign_image.BGR_INT16:
            ros_encoding = "bgr16"
            num_channels = 3
            octets_per_channel = 2
        elif format is ign_image.BGR_INT32:
            ros_encoding = "unknown"
            num_channels = 3
            octets_per_channel = 4
        elif format is ign_image.R_FLOAT16:
            ros_encoding = "16FC1"
            num_channels = 1
            octets_per_channel = 2
        elif format is ign_image.RGB_FLOAT16:
            ros_encoding = "16FC3"
            num_channels = 3
            octets_per_channel = 2
        elif format is ign_image.R_FLOAT32:
            ros_encoding = "32FC1"
            num_channels = 1
            octets_per_channel = 4
        elif format is ign_image.RGB_FLOAT32:
            ros_encoding = "32FC3"
            num_channels = 3
            octets_per_channel = 4
        elif format is ign_image.BAYER_RGGB8:
            ros_encoding = "bayer_rggb8"
            num_channels = 4
            octets_per_channel = 1
        elif format is ign_image.BAYER_BGGR8:
            ros_encoding = "bayer_bggr8"
            num_channels = 4
            octets_per_channel = 1
        elif format is ign_image.BAYER_GBRG8:
            ros_encoding = "bayer_gbrg8"
            num_channels = 4
            octets_per_channel = 1
        elif format is ign_image.BAYER_GRBG8:
            ros_encoding = "bayer_grbg8"
            num_channels = 4
            octets_per_channel = 1
        else:
            ros_encoding = "unknown"

        # step = width * num_channels * octets_per_channel
        # count = step * height
        return ros_encoding, num_channels, octets_per_channel


class SensorBridge(Node):

    def __init__(self):
        super().__init__('sensor_bridge')

        world_name = "omnicopter_runway"
        model_name = "omnicopter"

        self._sensors = {
            "camera_names" : [
                "camera",
            ],
            "depth_camera_names" : [
                "depth_camera",
            ],
            "rgbd_camera_names" : [
                "rgbd_camera/image",
                "rgbd_camera/depth_image",
            ],
            "thermal_camera_names" : [
                "thermal_camera",
            ]
        }

        # Timer
        timer_period = 0.5  # seconds
        self.timer = self.create_timer(timer_period, self.timer_callback)

        # ROS2 camera publishers
        self._camera_publishers = {}
        for sensor_name in self._sensors["camera_names"]:
            self._camera_publishers[sensor_name] = CameraPublishers(
                self, world_name, model_name, sensor_name)

        # ROS2 depth_camera publishers
        self._depth_camera_publishers = {}
        for sensor_name in self._sensors["depth_camera_names"]:
            self._depth_camera_publishers[sensor_name] = CameraPublishers(
                self, world_name, model_name, sensor_name)

        # ROS2 rgbd_camera publishers
        self._rgbd_camera_publishers = {}
        for sensor_name in self._sensors["rgbd_camera_names"]:
            self._rgbd_camera_publishers[sensor_name] = CameraPublishers(
                self, world_name, model_name, sensor_name)

        # ROS2 thermal_camera publishers
        self._thermal_camera_publishers = {}
        for sensor_name in self._sensors["thermal_camera_names"]:
            self._thermal_camera_publishers[sensor_name] = CameraPublishers(
                self, world_name, model_name, sensor_name)

        # Subscriptions (Ignition)
        self._ign_node = IgnNode()

        # subscription helper
        def subscribe(topic, callback, msg_type_name, opts):
          if self._ign_node.subscribe(topic, callback, msg_type_name, opts):
              self.get_logger().info(
                  "Subscribing to type [{}] on topic [{}]".format(
                      msg_type_name, topic))
          else:
              self.get_logger().error(
                  "Error subscribing to topic [{}]".format(topic))

        # Ignition camera subscribers
        self._camera_subscribers = {}
        for sensor_name in self._sensors["camera_names"]:
            self._camera_subscribers[sensor_name] = CameraSubscribers(
                self, world_name, model_name, sensor_name)

        # Ignition depth_camera subscribers
        self._depth_camera_subscribers = {}
        for sensor_name in self._sensors["depth_camera_names"]:
            self._depth_camera_subscribers[sensor_name] = CameraSubscribers(
                self, world_name, model_name, sensor_name)

        # Ignition rgbd_camera subscribers
        self._rgbd_camera_subscribers = {}
        for sensor_name in self._sensors["rgbd_camera_names"]:
            self._rgbd_camera_subscribers[sensor_name] = CameraSubscribers(
                self, world_name, model_name, sensor_name)

        # Ignition thermal_camera subscribers
        self._thermal_camera_subscribers = {}
        for sensor_name in self._sensors["thermal_camera_names"]:
            self._thermal_camera_subscribers[sensor_name] = CameraSubscribers(
                self, world_name, model_name, sensor_name)

    def timer_callback(self):
        # Publish last camera message
        for name in self._sensors["camera_names"]:
            sub = self._camera_subscribers[name]
            pub = self._camera_publishers[name]
            ros_image = self.convert_ign_to_ros(sub.last_image)
            pub.pub_image(ros_image)

        # Publish last depth_camera message
        for name in self._sensors["depth_camera_names"]:
            sub = self._depth_camera_subscribers[name]
            pub = self._depth_camera_publishers[name]
            ros_image = self.convert_ign_to_ros(sub.last_image)
            pub.pub_image(ros_image)

        # Publish last rgbd_camera message
        for name in self._sensors["rgbd_camera_names"]:
            sub = self._rgbd_camera_subscribers[name]
            pub = self._rgbd_camera_publishers[name]
            ros_image = self.convert_ign_to_ros(sub.last_image)
            pub.pub_image(ros_image)

        # Publish last thermal_camera message
        for name in self._sensors["thermal_camera_names"]:
            sub = self._thermal_camera_subscribers[name]
            pub = self._thermal_camera_publishers[name]
            ros_image = self.convert_ign_to_ros(sub.last_image)
            pub.pub_image(ros_image)

    def convert_ign_to_ros(self, image):
        stamp = builtin_interfaces.msg.Time(
            sec=image.header.stamp.sec,
            nanosec=image.header.stamp.nsec)

        encoding, num_channels, octets_per_channel = PixelFormatType.ign_to_ros(
            image.pixel_format_type)
        step = image.width * num_channels * octets_per_channel
        count = step * image.height

        # ROS2 sensor_msgs.msg.Image
        ros_image = Image()
        ros_image.header.stamp = stamp
        ros_image.header.frame_id = "sensor_link"
        ros_image.height = image.height
        ros_image.width = image.width
        ros_image.step = step
        ros_image.encoding = encoding
        ros_image.is_bigendian = False
        ros_image.data = image.data

        return ros_image


def main(args=None):
    rclpy.init(args=args)

    sensor_bridge = SensorBridge()

    rclpy.spin(sensor_bridge)

    # Destroy the node explicitly
    # (optional - otherwise it will be done automatically
    # when the garbage collector destroys the node object)
    sensor_bridge.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()